p.welcome{
  margin-top: 30px;
  margin-left: 40px;
  font-size: 16px;
}

.usecase{
  margin-top: 100px;
}

.file-upload {
  display: inline;
}

.download {
  display: inline;
}

html,
body {
  height: 100%;
  margin: 0
}
