# Generated by Django 3.2.10 on 2022-01-27 10:45

import django.core.files.storage
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('demo', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='Downloader',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('first_name', models.CharField(max_length=20)),
                ('last_name', models.CharField(max_length=20)),
                ('email', models.EmailField(max_length=50)),
                ('affiliation', models.CharField(max_length=100)),
                ('funder', models.CharField(blank=True, default='', max_length=100)),
                ('use_case', models.TextField(max_length=200)),
            ],
        ),
        migrations.CreateModel(
            name='MedcatModel',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('model_name', models.CharField(max_length=20, unique=True)),
                ('model_file', models.FileField(storage=django.core.files.storage.FileSystemStorage(location='/medcat_data'), upload_to='')),
                ('model_display_name', models.CharField(max_length=50)),
            ],
        ),
    ]
